#include "student.h"
#include <algorithm>
#include <numeric>
#include <iostream>
#include <utility>
#include <vector>
// for std::function
#include <functional>
#include <set>

using namespace std;

struct SqrtError {};

// Compute square root using Newton's method
double Sqrt(double a) {
  if (a < 0) {
    throw SqrtError{};
  }
  
  if (a == 0) {
    return 0;
  }
  
  // calculate 2^(log₂ a) as the initial approximation
  double x = a;
  for (double t = a; t > 1; t /= 4, x /= 2) {}
  for (double t = a; t < 1; t *= 4, x *= 2) {}
  
  // 8 iterations are enough for doubles at this point
  for (int i = 0; i < 8; ++i) {
    // xᵢ₊₁ = xᵢ - f(xᵢ)/f'(xᵢ)
    //
    // where f(x) = x² - a, so the roots are +/- √a
    //       f'(x) = 2x
    x = (x + (a / x) ) / 2;
  }
  return x;
}

// other math functions
double Abs(double x) {
  if (x == 0) return +0.0;
  if (x < 0) return -x;
  return x;
}

double Square(double x) {
  return x * x;
}

int Round(double x) {
  return (int)x;
}

template<class It>
void Print(It begin, It end) {
  for (; begin != end; ++begin) {
    cout << *begin << '\t';
  }
  cout << '\n';
}

template<class Vector>
void PrintCollection(const string & name, Vector v) {
  cout << name << ":\t";
  Print(v.begin(), v.end());
}

// typedef vector<Student> Roster;
// using Roster = vector<Student>;

int Length(const string & name) {
  return name.size();
}

template<typename In, typename Out>
vector<Out> Map(vector<In> v, Out (f)(In)) {
  vector<Out> result;
  for (auto & x : v) {
    result.push_back(f(x));
  }
  return result;
}

int main(int argc, char * argv[]) {
  vector<double> v{-2.718, -1, 0, 1, 2, 3, 4, 3.14};
  vector<double> v2{0, 1, 2, 3, 4, 3.14, 36, 64, 1e32};

  // whence functions?
  
  vector<double> squares = Map(v, Square);
  vector<double> absolutes = Map(v, Abs);
  vector<int> converted_to_ints = Map(v, Round);
  vector<double> square_roots(v2.size());
  std::transform(v2.begin(), v2.end(), square_roots.begin(), Sqrt);
  // ---- map

  // PrintCollection("squares", squares);
  // PrintCollection("vector 1", v);
  // PrintCollection("absolutes", absolutes);
  // PrintCollection("converted ints", converted_to_ints);
  // PrintCollection("vector 2", v2);
  // PrintCollection("square roots", square_roots);
  // PrintCollection("square roots 2", sqrt2);

  // ---- reduce

  vector<int> firstN{1, 2, 3, 4, 5, 6, 7, 8, 9, 10};
  int fact = std::reduce(firstN.begin(), firstN.end(), 1,
                    [](int a, int b) { return a * b; }
                    );

  cout << "10! = " << fact << "\n";

  // ---- transform_reduce, std library fns
  vector<string> names{"Mehmet", "Alice", "John"};
  size_t buf_len = 0;

  buf_len = std::transform_reduce(names.begin(), names.end(), 0, std::plus(), Length);
  
  cout << "total length: " << buf_len << "\n";
  
  // find, find_if
  auto i = std::find_if(names.begin(), names.end(), [](auto& name) {
    return name.size() > 0 && name[0] == 'A';
  });
  if (i != names.end()) {
    *i = "Bob";
  }

  PrintCollection("names", names);

  // TODO: one of
  //  - find a specific element
  //  - is everything positive/even/etc.

  vector<Student> roster{
    Student("Mehmet", "Emre", "CS"),
    Student("Arthur", "Murray", "Dance"),
    Student("Charles", "Darwin", "Biology"),
    Student("Alan", "Turing", "CS"),
  };

  // TODO:
  // - sort the roster so we can do binary search
  // - sort the roster by major
  // - sort the roster by first name/last name
  // - get only CS majors (drop others)

  for (auto& student : roster) {
    cout << student.first_name << " " << student.last_name << "\n";
  }

  std::sort(roster.begin(), roster.end(),
            [](Student& a, Student& b) {
              return a.first_name < b.first_name;
            }
            );

  cout << "sorted:\n";
  
  for (auto& student : roster) {
    cout << student.first_name << " " << student.last_name << "\n";
  }

  auto new_end =std::remove_if(roster.begin(),
                 roster.end(),
                 [](const Student& s) {
                   return s.major != "CS";
                 });
  roster.erase(new_end, roster.end());

  cout << "CS only:\n";

  for (auto& student : roster) {
    cout << student.first_name << " " << student.last_name << "\n";
  }
  
  return 0;
}
