#include <algorithm>
#include <cctype>
#include <iostream>
#include <fstream>
#include <sstream>
#include <stdexcept>
#include <utility>
#include <vector>
// for std::function
#include <functional>
#include <unordered_set>
#include <unordered_map>

using namespace std;

template<class It>
void Print(It begin, It end) {
  for (; begin != end; ++begin) {
    cout << *begin << '\t';
  }
  cout << '\n';
}

template<class Vector>
void PrintCollection(const string & name, Vector v) {
  cout << name << ":\t";
  Print(v.begin(), v.end());
}

std::string MakeValidWord(const std::string& word) {
  auto is_letter = [](char c) -> bool { return std::isalpha(c); };
  auto begin = std::find_if(word.begin(), word.end(), is_letter);
  auto end = std::find_if(word.rbegin(), word.rend(), is_letter).base();

  if (begin == word.end()) {
    return "";
  }

  ostringstream result;
  for (; begin != end; ++begin) {
    result.put(std::tolower(*begin));
  }
  
  return result.str();
}

class MyLambda {
  unordered_map<string, int> &freq;

  void operator() (const std::string& word) { freq[word]++; }
};

void ForAllWords(string_view file_name,
                 std::function<void (const std::string&)> f
                 ) {
  unordered_set<string> CommonWords {
    "the", "a", "i", "you", "to", "from", "of", "my", "your", "their", "he", "she", "it", "him", "her", "his", "is", "are", "were", "will",
    "and", "in", "not", "this", "with", "but", "for", "be", "me", "was", "as", "this", "that", "so", "have", "has", "had", "what", "where", "why", "how",
    "do", "no", "we", "on", "by", "our", "they", "them", "if", "or", "shall", "thou", "thee", "o"
  };
  
  fstream file(file_name);
  if (!file) {
    throw std::runtime_error("file does not exist: " + string(file_name));
  }

  string word;
  while (file && ! file.eof()) {
    file >> word;
    word = MakeValidWord(word);
    if (word == "" || CommonWords.count(word)) {
      continue;
    }
    f(word);
  }
}

int main(int argc, char * argv[]) {
  if (argc != 2) {
    cerr << "usage: " << argv[0] << " <text-file>\n";
    return 1;
  }

  // ForAllWords(argv[1], [](const std::string& word) {
  //   cout << word << "\n";
  // });

  // order by frequency
  unordered_map<string, int> freq;
  int amount = 1;

  ForAllWords(argv[1], [=, &freq](const std::string& word) {
    freq[word] += amount;
  });
  
  vector<pair<string, int>> word_stat(freq.begin(), freq.end());
  std::sort(word_stat.begin(), word_stat.end(), [](auto& a, auto& b) { return a.second < b.second; });

  int count = 10;
  for (auto i = word_stat.rbegin(); count > 0 && i != word_stat.rend(); count--, i++) {
    cout << setw(15) << i->first << ": " << i->second << "\n";
  }
  
  return 0;
}
